/*
 * Encoding: UTF-8
 * File:     proxyhttp-tester.cc
 * Project:  ISA - Simple HTTP proxy server
 * Date:     20.10.2007
 * Author:   Petr Zemek, xzemek02@stud.fit.vutbr.cz
 */

#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include "proxyhttp.h"

using namespace std;
using namespace CppUnit;
using namespace ProxyHTTP;

//! CommandLineArguments testing suite.
class CommandLineArgumentsTest : public CppUnit::TestFixture {
private:
	CPPUNIT_TEST_SUITE(CommandLineArgumentsTest);
	CPPUNIT_TEST(testObjectCreationNoArguments);
	CPPUNIT_TEST(testObjectCreationOneArgument);
	CPPUNIT_TEST(testObjectCreationThreeArguments);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsLocalPortInvalid);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsLocalPortTooLow);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsLocalPortTooHigh);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsHostPortEmpty);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsHostPortInvalid);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsHostColonButNoPort);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsHostMissing);
	CPPUNIT_TEST(testObjectCreationTwoArgumentsEverythingSet);
	CPPUNIT_TEST_SUITE_END();

public:
	void scenarioInvalidArguments(int argc, const char **argv,
		const string &exceptionMessage)
	{
		try {
			CommandLineArguments args(argc, argv);
		}
		catch (const CommandLineArgumentsException &e) {
			CPPUNIT_ASSERT(e.what() == exceptionMessage);
			return;
		}
	}

	void scenarioValidArguments(int argc, const char **argv,
		unsigned int localPort, string hostAddress, unsigned int hostPort)
	{
		CommandLineArguments args(argc, argv);

		CPPUNIT_ASSERT(args.localPort() == localPort);
		CPPUNIT_ASSERT(args.hostAddress() == hostAddress);
		CPPUNIT_ASSERT(args.hostPort() == hostPort);
	}

	void testObjectCreationNoArguments()
	{
		int argc = 1;
		const char *argv[] = {
			"tester",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid arguments count."));
	}

	void testObjectCreationOneArgument()
	{
		int argc = 2;
		const char *argv[] = {
			"tester",
			"80",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid arguments count."));
	}

	void testObjectCreationThreeArguments()
	{
		int argc = 2;
		const char *argv[] = {
			"tester",
			"80",
			"www.fit.vutbr.cz",
			"unknown",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid arguments count."));
	}

	void testObjectCreationTwoArgumentsLocalPortInvalid()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"abcd",
			"www.fit.vutbr.cz",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid port number: abcd."));
	}

	void testObjectCreationTwoArgumentsLocalPortTooLow()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"0",
			"www.fit.vutbr.cz",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid port number: 0."));
	}

	void testObjectCreationTwoArgumentsLocalPortTooHigh()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"65536",
			"www.fit.vutbr.cz",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid port number: 65536."));
	}

	void testObjectCreationTwoArgumentsHostPortEmpty()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"8080",
			"www.fit.vutbr.cz",
		};

		scenarioValidArguments(argc, argv, 8080, string("www.fit.vutbr.cz"), 80);
	}

	void testObjectCreationTwoArgumentsHostPortInvalid()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"8080",
			"www.fit.vutbr.cz:AA",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid port number: AA."));
	}

	void testObjectCreationTwoArgumentsHostColonButNoPort()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"6666",
			"www.fit.vutbr.cz:",
		};

		scenarioInvalidArguments(argc, argv, string("Invalid port number: ."));
	}

	void testObjectCreationTwoArgumentsHostMissing()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"11000",
			":80",
		};

		scenarioInvalidArguments(argc, argv, string("Host address is missing."));
	}

	void testObjectCreationTwoArgumentsEverythingSet()
	{
		int argc = 3;
		const char *argv[] = {
			"tester",
			"8080",
			"www.fit.vutbr.cz:8080",
		};

		scenarioValidArguments(argc, argv, 8080, string("www.fit.vutbr.cz"), 8080);
	}
};

CPPUNIT_TEST_SUITE_REGISTRATION(CommandLineArgumentsTest);

//! HttpRequest testing suite.
class HttpRequestTest : public CppUnit::TestFixture {
private:
	CPPUNIT_TEST_SUITE(HttpRequestTest);
	CPPUNIT_TEST(testRequestCreationFromEmptyString);
	CPPUNIT_TEST(testRequestCreationValidHttp11Request);
	CPPUNIT_TEST(testHttpVersionRetrievalHttp10);
	CPPUNIT_TEST(testHttpVersionRetrievalHttp11);
	CPPUNIT_TEST(testHttpVersionRetrievalUnknown);
	CPPUNIT_TEST(testHttpRequestTypeRetrievalGet);
	CPPUNIT_TEST(testHttpRequestTypeRetrievalPost);
	CPPUNIT_TEST(testHttpRequestTypeRetrievalHead);
	CPPUNIT_TEST(testRequestCompletenessCompleteRequest);
	CPPUNIT_TEST(testRequestCompletenessIncompleteRequest);
	CPPUNIT_TEST(testGetHeaderValueEmptyHeaderName);
	CPPUNIT_TEST(testGetHeaderValueExistingHeaderCaseInsensitivity);
	CPPUNIT_TEST(testGetHeaderValueExistingHeaderNoFollowingHeader);
	CPPUNIT_TEST(testGetHeaderValueExistingHeaderFollowingHeaderExists);
	CPPUNIT_TEST(testGetHeaderValueNonExistingHeader);
	CPPUNIT_TEST(testObjectToGetfileNameWrongMethod);
	CPPUNIT_TEST(testObjectToGetfileNameEmptyRequest);
	CPPUNIT_TEST(testObjectToGetfileNameInvalidRequest);
	CPPUNIT_TEST(testObjectToGetfileNameRootDirectory);
	CPPUNIT_TEST(testObjectToGetfileNameNoSubdirectories);
	CPPUNIT_TEST(testObjectToGetfileNameWithSubdirectories);
	CPPUNIT_TEST(testObjectToGetfileNameWithSubdirectoriesEndsWithSubdirectory);
	CPPUNIT_TEST(testObjectToGetfileNameWithArgument);
	CPPUNIT_TEST(testObjectToGetfileNameWithArguments);
	CPPUNIT_TEST(testChangeHeaderValueHost);
	CPPUNIT_TEST(testChangeHeaderValueHostCaseInsensitivity);
	CPPUNIT_TEST(testChangeHeaderValueEncoding);
	CPPUNIT_TEST(testChangeHeaderValueEncodingCaseInsensitivity);
	CPPUNIT_TEST(testChangeHeaderValueNonExistingHeaderName);
	CPPUNIT_TEST(testChangeHeaderValueEmptyHeaderName);
	CPPUNIT_TEST(testChangeHeaderValueEmptyNewHeaderValue);
	CPPUNIT_TEST(testConnectionHasToBePersistentHttp10);
	CPPUNIT_TEST(testConnectionHasToBePersistentHttp11ConnectionClose);
	CPPUNIT_TEST(testConnectionHasToBePersistentHttp11NoConnectionClose);
	CPPUNIT_TEST(testIsCompleteGetRequestComplete);
	CPPUNIT_TEST(testIsCompleteGetRequestNotComplete);
	CPPUNIT_TEST(testIsCompletePostRequestComplete);
	CPPUNIT_TEST(testIsCompletePostRequestNotComplete);
	CPPUNIT_TEST_SUITE_END();

public:
	void scenarioRequestCreation(const string &requestMessage)
	{
		HttpRequest request(requestMessage);

		CPPUNIT_ASSERT(request.rawMessage() == requestMessage);
	}

	void testRequestCreationFromEmptyString()
	{
		scenarioRequestCreation(string(""));
	}

	void testRequestCreationValidHttp11Request()
	{
		scenarioRequestCreation(string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"));
	}

	void scenarioHttpVersionRetrieval(const string &httpVersion,
		const string &referenceHttpVersion)
	{
		string requestMessage(string("GET / ") + httpVersion +
			string("\r\nHost: localhost\r\n\r\n"));
		HttpRequest request(requestMessage);

		CPPUNIT_ASSERT(request.httpVersion() == referenceHttpVersion);
	}

	void testHttpVersionRetrievalHttp10()
	{
		scenarioHttpVersionRetrieval(HttpMessage::HttpVersion10Marker,
			HttpMessage::HttpVersion10Marker);
	}

	void testHttpVersionRetrievalHttp11()
	{
		scenarioHttpVersionRetrieval(HttpMessage::HttpVersion10Marker,
			HttpMessage::HttpVersion10Marker);
	}

	void testHttpVersionRetrievalUnknown()
	{
		scenarioHttpVersionRetrieval(string("$#%DSF"), string(""));
	}

	void scenarioHttpRequestTypeRetrieval(const string &httpRequestType,
		const string &referenceHttpRequestType)
	{
		string requestMessage(httpRequestType +
			string(" / HTTP/1.1\r\nHost: localhost\r\n\r\n"));
		HttpRequest request(requestMessage);

		CPPUNIT_ASSERT(request.httpRequestType() == referenceHttpRequestType);
	}

	void testHttpRequestTypeRetrievalGet()
	{
		scenarioHttpRequestTypeRetrieval(HttpRequest::HttpRequestTypeGet,
			HttpRequest::HttpRequestTypeGet);
	}

	void testHttpRequestTypeRetrievalPost()
	{
		scenarioHttpRequestTypeRetrieval(string("POST"), string("POST"));
	}

	void testHttpRequestTypeRetrievalHead()
	{
		scenarioHttpRequestTypeRetrieval(string("HEAD"), string("HEAD"));
	}

	void scenarioTestRequestCompleteness(const string &requestMessage,
		bool shouldBeComplete)
	{
		HttpRequest request(requestMessage);

		CPPUNIT_ASSERT(request.isComplete() == shouldBeComplete);
	}

	void testRequestCompletenessCompleteRequest()
	{
		scenarioTestRequestCompleteness(
			string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"), true);
	}

	void testRequestCompletenessIncompleteRequest()
	{
		scenarioTestRequestCompleteness(
			string("GET / HTTP/1.1\r\nHost:"), false);
	}

	void scenarioGetHeaderValue(const string &requestMessage,
		const string &headerName, const string &referenceValue)
	{
		HttpRequest request(requestMessage);

		CPPUNIT_ASSERT(request.getHeaderValue(headerName) == referenceValue);
	}

	void testGetHeaderValueEmptyHeaderName()
	{
		scenarioGetHeaderValue(string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string(""), string(""));
	}

	void testGetHeaderValueExistingHeaderCaseInsensitivity()
	{
		scenarioGetHeaderValue(string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("hoST"), string("localhost"));
	}

	void testGetHeaderValueExistingHeaderNoFollowingHeader()
	{
		scenarioGetHeaderValue(string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("Host"), string("localhost"));
	}

	void testGetHeaderValueExistingHeaderFollowingHeaderExists()
	{
		scenarioGetHeaderValue(string("GET / HTTP/1.1\r\nHost: localhost\r\nContent-Length: 1024\r\n\r\n"),
			string("Host"), string("localhost"));
	}

	void testGetHeaderValueNonExistingHeader()
	{
		scenarioGetHeaderValue(string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("Content-Length"), string(""));
	}

	void scenarioObjectToGetfileName(const string &requestMessage,
		const string &referenceObjectFileName)
	{
		HttpRequest request(requestMessage);

		CPPUNIT_ASSERT(request.objectToGetfileName() == referenceObjectFileName);
	}

	void testObjectToGetfileNameWrongMethod()
	{
		scenarioObjectToGetfileName(string(
			"HEAD / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string(""));
	}

	void testObjectToGetfileNameEmptyRequest()
	{
		scenarioObjectToGetfileName(string(""), string(""));
	}

	void testObjectToGetfileNameInvalidRequest()
	{
		scenarioObjectToGetfileName(string("GET /"), string(""));
	}

	void testObjectToGetfileNameRootDirectory()
	{
		scenarioObjectToGetfileName(string(
			"GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("index.html"));
	}

	void testObjectToGetfileNameNoSubdirectories()
	{
		scenarioObjectToGetfileName(string(
			"GET /index2.html HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("index2.html"));
	}

	void testObjectToGetfileNameWithSubdirectories()
	{
		scenarioObjectToGetfileName(string(
			"GET /path/i.php HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("i.php"));
	}

	void testObjectToGetfileNameWithSubdirectoriesEndsWithSubdirectory()
	{
		scenarioObjectToGetfileName(string(
			"GET /path/subdir/ HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("index.html"));
	}

	void testObjectToGetfileNameWithArgument()
	{
		scenarioObjectToGetfileName(string(
			"GET /path/test.php?param=value HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("test.php"));
	}

	void testObjectToGetfileNameWithArguments()
	{
		scenarioObjectToGetfileName(string(
			"GET /path/index.php?m=main&p=introduction HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("index.php"));
	}

	void scenarioTestChangeHeaderValue(const HttpRequest &originalHttpRequest,
		const string &headerName, const string &newHeaderValue,
		bool shouldSucceed, const HttpRequest &referenceHttpRequest)
	{
		HttpRequest originalRequestCopy(originalHttpRequest);
		bool succeeded = originalRequestCopy.changeHeaderValue(headerName, newHeaderValue);

		CPPUNIT_ASSERT(succeeded == shouldSucceed);
		CPPUNIT_ASSERT(originalRequestCopy.rawMessage() ==
			referenceHttpRequest.rawMessage());
	}

	void testChangeHeaderValueHost()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("Host"), string("www.fit.vutbr.cz"), true,
			HttpRequest("GET / HTTP/1.1\r\nHost: www.fit.vutbr.cz\r\n\r\n")
		);
	}

	void testChangeHeaderValueHostCaseInsensitivity()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("HOsT"), string("www.fit.vutbr.cz"), true,
			HttpRequest("GET / HTTP/1.1\r\nHost: www.fit.vutbr.cz\r\n\r\n")
		);
	}

	void testChangeHeaderValueEncoding()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nEncoding: utf8\r\n\r\n"),
			string("Encoding"), string("latin2"), true,
			HttpRequest("GET / HTTP/1.1\r\nEncoding: latin2\r\n\r\n")
		);
	}

	void testChangeHeaderValueEncodingCaseInsensitivity()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nEncodinG: utf8\r\n\r\n"),
			string("EncoDing"), string("latin2"), true,
			HttpRequest("GET / HTTP/1.1\r\nEncodinG: latin2\r\n\r\n")
		);
	}

	void testChangeHeaderValueNonExistingHeaderName()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string("Encoding"), string("latin2"), false,
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n")
		);
	}

	void testChangeHeaderValueEmptyHeaderName()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string(""), string("www.fit.vutbr.cz"), false,
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n")
		);
	}

	void testChangeHeaderValueEmptyNewHeaderValue()
	{
		scenarioTestChangeHeaderValue(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"),
			string(""), string(""), false,
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n")
		);
	}

	void scenarioTestConnectionHasToBePersistent(const HttpRequest &request,
		bool shouldBePersistent)
	{
		CPPUNIT_ASSERT(request.connectionHasToBePersistent() == shouldBePersistent);
	}

	void testConnectionHasToBePersistentHttp10()
	{
		scenarioTestConnectionHasToBePersistent(
			HttpRequest("GET / HTTP/1.0\r\n\r\n"), false);
	}

	void testConnectionHasToBePersistentHttp11ConnectionClose()
	{
		scenarioTestConnectionHasToBePersistent(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\nConnection: close\r\n\r\n"), false);
	}

	void testConnectionHasToBePersistentHttp11NoConnectionClose()
	{
		scenarioTestConnectionHasToBePersistent(
			HttpRequest("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n"), true);
	}

	void scenarioTestIsCompleteRequest(const HttpRequest &request, bool shouldBeComplete)
	{
		CPPUNIT_ASSERT(request.isComplete() == shouldBeComplete);
	}

	void testIsCompleteGetRequestComplete()
	{
		scenarioTestIsCompleteRequest(
			HttpRequest(string("GET / HTTP/1.1\r\nHost: localhost\r\n\r\n")),
			true);
	}

	void testIsCompleteGetRequestNotComplete()
	{
		scenarioTestIsCompleteRequest(
			HttpRequest(string("GET / HTTP/1.1\r\nHost: l")),
			false);
	}

	void testIsCompletePostRequestComplete()
	{
		scenarioTestIsCompleteRequest(
			HttpRequest(string("POST /test.html HTTP/1.1\r\n"
				"Host: localhost\r\n"
				"Content-Length: 10\r\n"
				"\r\n"
				"abcdefghtr")),
			true);
	}

	void testIsCompletePostRequestNotComplete()
	{
		scenarioTestIsCompleteRequest(
			HttpRequest(string("POST /test.html HTTP/1.1\r\n"
				"Host: localhost\r\n"
				"Content-Length: 40\r\n"
				"\r\n"
				"abcdefghtr")),
			false);
	}
};

CPPUNIT_TEST_SUITE_REGISTRATION(HttpRequestTest);

//! HttpResponse testing suite.
class HttpResponseTest : public CppUnit::TestFixture {
private:
	CPPUNIT_TEST_SUITE(HttpResponseTest);
	CPPUNIT_TEST(testResponseCreationFromEmptyString);
	CPPUNIT_TEST(testResponseCreationValidHttp11Response);
	CPPUNIT_TEST(testRawMessageBodyEmptyBody);
	CPPUNIT_TEST(testRawMessageBodyNonEmptyBody);
	CPPUNIT_TEST(testHttpVersionRetrievalHttp10);
	CPPUNIT_TEST(testHttpVersionRetrievalHttp11);
	CPPUNIT_TEST(testHttpVersionRetrievalUnknown);
	CPPUNIT_TEST(testResponseCompletenessContentLengthCompleteResponse);
	CPPUNIT_TEST(testResponseCompletenessContentLengthIncompleteResponse);
	CPPUNIT_TEST(testResponseCompletenessContentLengthEmptyResponse);
	CPPUNIT_TEST(testResponseCompletenessTransferEncodingChunkedCompleteResponse);
	CPPUNIT_TEST(testResponseCompletenessTransferEncodingChunkedIncompleteResponse);
	CPPUNIT_TEST(testResponseCompletenessTransferEncodingChunkedEmptyResponse);
	CPPUNIT_TEST(testResponseCode200Ok);
	CPPUNIT_TEST(testResponseCode304NotModified);
	CPPUNIT_TEST(testResponseCodeEmptyResponseMessage);
	CPPUNIT_TEST(testResponseCodeInvalidCode);
	CPPUNIT_TEST_SUITE_END();

public:
	void scenarioResponseCreation(const string &responseMessage)
	{
		HttpResponse response(responseMessage);

		CPPUNIT_ASSERT(response.rawMessage() == responseMessage);
	}

	void testResponseCreationFromEmptyString()
	{
		scenarioResponseCreation(string(""));
	}

	void testResponseCreationValidHttp11Response()
	{
		scenarioResponseCreation(string("HTTP/1.1 200 OK\r\nText"));
	}

	void scenarioRawMessageBody(const string &rawMessage,
		const string &referenceRawMessageBody)
	{
		HttpResponse response(rawMessage);

		CPPUNIT_ASSERT(response.rawMessageBody() == referenceRawMessageBody);
	}

	void testRawMessageBodyEmptyBody()
	{
		scenarioRawMessageBody(string("HTTP/1.1 200 OK\r\n\r\n"), string(""));
	}

	void testRawMessageBodyNonEmptyBody()
	{
		scenarioRawMessageBody(string("HTTP/1.1 200 OK\r\n\r\nbody"),
			string("body"));
	}

	void scenarioHttpVersionRetrieval(const string &httpVersion,
		const string &referenceHttpVersion)
	{
		string responseMessage(string(httpVersion + string(" 200 OK\r\nText")));
		HttpResponse response(responseMessage);

		CPPUNIT_ASSERT(response.httpVersion() == referenceHttpVersion);
	}

	void testHttpVersionRetrievalHttp10()
	{
		scenarioHttpVersionRetrieval(HttpMessage::HttpVersion10Marker,
			HttpMessage::HttpVersion10Marker);
	}

	void testHttpVersionRetrievalHttp11()
	{
		scenarioHttpVersionRetrieval(HttpMessage::HttpVersion10Marker,
			HttpMessage::HttpVersion10Marker);
	}

	void testHttpVersionRetrievalUnknown()
	{
		scenarioHttpVersionRetrieval(string("$#%DSF"), string(""));
	}

	void scenarioTestResponseCompleteness(const string &responseMessage,
		bool shouldBeComplete)
	{
		HttpResponse response(responseMessage);

		CPPUNIT_ASSERT(response.isComplete() == shouldBeComplete);
	}

	void testResponseCompletenessContentLengthCompleteResponse()
	{
		scenarioTestResponseCompleteness(string(
			"HTTP/1.1 200 OK\r\n"
			"Content-Length: 4\r\n"
			"\r\n"
			"Test"), true);
	}

	void testResponseCompletenessContentLengthIncompleteResponse()
	{
		scenarioTestResponseCompleteness(string(
			"HTTP/1.1 200 OK\r\n"
			"Content-Length: 20\r\n"
			"\r\n"
			"Test"), false);
	}

	void testResponseCompletenessContentLengthEmptyResponse()
	{
		scenarioTestResponseCompleteness(string(""), false);
	}

	void testResponseCompletenessTransferEncodingChunkedCompleteResponse()
	{
		scenarioTestResponseCompleteness(string(
			"HTTP/1.1 200 OK\r\n"
			"Transfer-Encoding: chunked\r\n"
			"\r\n"
			"ffb\r\n"
			"asafdsgsg\r\n" // Well, there should be more text, but it doesn't matter
							// (I'm testing only the end marker availability)
			"0\r\n\r\n"), true);
	}

	void testResponseCompletenessTransferEncodingChunkedIncompleteResponse()
	{
		scenarioTestResponseCompleteness(string(
			"HTTP/1.1 200 OK\r\n"
			"Transfer-Encoding: chunked\r\n"
			"\r\n"
			"eca\r\n"
			"abcdfdg"), false);
	}

	void testResponseCompletenessTransferEncodingChunkedEmptyResponse()
	{
		scenarioTestResponseCompleteness(string(""), false);
	}

	void scenarioTestResponseCode(const string &responseMessage,
		const unsigned int referenceResponseCode)
	{
		HttpResponse response(responseMessage);

		CPPUNIT_ASSERT(response.responseCode() == referenceResponseCode);
	}

	void testResponseCode200Ok()
	{
		scenarioTestResponseCode(string(
			"HTTP/1.1 200 OK\r\n"
			"Content-Length: 5\r\n"
			"\r\n"
			"Test."), 200);
	}

	void testResponseCode304NotModified()
	{
		scenarioTestResponseCode(string(
			"HTTP/1.1 304 Not Modified\r\n"
			"\r\n"), 304);
	}

	void testResponseCodeEmptyResponseMessage()
	{
		scenarioTestResponseCode(string(""), 0);
	}

	void testResponseCodeInvalidCode()
	{
		scenarioTestResponseCode(string(
			"HTTP/1.1 abc Oops\r\n"
			"\r\n"), 0);
	}
};

CPPUNIT_TEST_SUITE_REGISTRATION(HttpResponseTest);

//! Server testing suite.
class ServerTest : public CppUnit::TestFixture {
private:
	CPPUNIT_TEST_SUITE(ServerTest);
 	CPPUNIT_TEST(testConnectionByHostnameToRemoteHostAndDisconnection);
 	CPPUNIT_TEST(testConnectionByIPAddressToRemoteHostAndDisconnection);
 	CPPUNIT_TEST(testConnectionToAlreadyConnectedRemoteHost);
 	CPPUNIT_TEST(testLoadForbiddenInfoListNoInfo);
 	CPPUNIT_TEST(testLoadForbiddenInfoListMissingClientAddress);
 	CPPUNIT_TEST(testLoadForbiddenInfoListMissingObjectFileName);
 	CPPUNIT_TEST(testLoadForbiddenInfoListNoColon);
 	CPPUNIT_TEST(testLoadForbiddenInfoListOneRecord);
 	CPPUNIT_TEST(testLoadForbiddenInfoListOneRecordDnsNameInsteadOfAddress);
 	CPPUNIT_TEST(testLoadForbiddenInfoListTwoRecords);
 	CPPUNIT_TEST(testLoadForbiddenInfoListTwoRecordsEndOfLineInTheEndOfTheSource);
 	CPPUNIT_TEST(testLoadForbiddenInfoListFiveRecords);
 	CPPUNIT_TEST(testIsForbiddenEmptyClientAddress);
 	CPPUNIT_TEST(testIsForbiddenEmptyObjectToGetFileName);
 	CPPUNIT_TEST(testIsForbiddenIncludesOnlyAPartNotWholeFileName);
 	CPPUNIT_TEST(testIsForbiddenNotForbiddenClientAddress);
 	CPPUNIT_TEST(testIsForbiddenNotForbiddenObjectToGetFileName);
	CPPUNIT_TEST_SUITE_END();

	typedef pair<string, string> ForbiddenInfo;
	typedef vector<ForbiddenInfo> ForbiddenInfoList;

public:
	void testConnectionByHostnameToRemoteHostAndDisconnection()
	{
		auto_ptr<Server> server(new Server(40657));
  		auto_ptr<ConnectionToHost> connectionToHost = server->connectToHost("localhost", 80);
	}

	void testConnectionByIPAddressToRemoteHostAndDisconnection()
	{
		auto_ptr<Server> server(new Server(40657));
  		auto_ptr<ConnectionToHost> connectionToHost = server->connectToHost("127.0.0.1", 80);
	}

	void testConnectionToAlreadyConnectedRemoteHost()
	{
		auto_ptr<Server> server(new Server(40657));
  		auto_ptr<ConnectionToHost> connectionToHost1 = server->connectToHost("127.0.0.1", 80);
  		auto_ptr<ConnectionToHost> connectionToHost2 = server->connectToHost("localhost", 80);
	}

	void scenarioTestLoadForbiddenInfoListNoInfo(const string &rawForbiddenInfo,
		const ForbiddenInfoList	&referenceForbiddenInfoList)
	{
		auto_ptr<Server> server(new Server(40658));
		server->loadForbiddenInfoList(rawForbiddenInfo);

		for (ForbiddenInfoList::const_iterator item = referenceForbiddenInfoList.begin();
			 item != referenceForbiddenInfoList.end(); ++item) {
			CPPUNIT_ASSERT(server->isForbidden(item->first, item->second));
		}
	}

	void testLoadForbiddenInfoListNoInfo()
	{
		scenarioTestLoadForbiddenInfoListNoInfo(string(""),
			ForbiddenInfoList());
	}

	void testLoadForbiddenInfoListMissingClientAddress()
	{
		try {
			scenarioTestLoadForbiddenInfoListNoInfo(string(":a.htm"),
				ForbiddenInfoList());
			CPPUNIT_ASSERT(false && "Should never get there.");
		}
		catch (const RuntimeException &e) {
			// OK, info list was invalid
			CPPUNIT_ASSERT(e.what() == string(
				"Data in the 'forbidden' file have invalid format."));
		}
	}

	void testLoadForbiddenInfoListMissingObjectFileName()
	{
		try {
			scenarioTestLoadForbiddenInfoListNoInfo(string("127.0.0.1:"),
				ForbiddenInfoList());
			CPPUNIT_ASSERT(false && "Should never get there.");
		}
		catch (const RuntimeException &e) {
			// OK, info list was invalid
			CPPUNIT_ASSERT(e.what() == string(
				"Data in the 'forbidden' file have invalid format."));
		}
	}

	void testLoadForbiddenInfoListNoColon()
	{
		try {
			scenarioTestLoadForbiddenInfoListNoInfo(string("127.0.0.1abcd.htm"),
				ForbiddenInfoList());
			CPPUNIT_ASSERT(false && "Should never get there.");
		}
		catch (const RuntimeException &e) {
			// OK, info list was invalid
			CPPUNIT_ASSERT(e.what() == string(
				"Data in the 'forbidden' file have invalid format."));
		}
	}

	void testLoadForbiddenInfoListOneRecord()
	{
		ForbiddenInfoList forbiddenInfoList;
		forbiddenInfoList.push_back(ForbiddenInfo("127.0.0.1", "index.html"));
		scenarioTestLoadForbiddenInfoListNoInfo(string("127.0.0.1:index.html"),
			forbiddenInfoList);
	}

	void testLoadForbiddenInfoListOneRecordDnsNameInsteadOfAddress()
	{
		ForbiddenInfoList forbiddenInfoList;
		forbiddenInfoList.push_back(ForbiddenInfo("www.proxy.com", "index.html"));
		scenarioTestLoadForbiddenInfoListNoInfo(string("www.proxy.com:index.html"),
			forbiddenInfoList);
	}

	void testLoadForbiddenInfoListTwoRecords()
	{
		ForbiddenInfoList forbiddenInfoList;
		forbiddenInfoList.push_back(ForbiddenInfo("127.0.0.1", "index.html"));
		forbiddenInfoList.push_back(ForbiddenInfo("167.9.11.2", "style.css"));
		scenarioTestLoadForbiddenInfoListNoInfo(
			string("127.0.0.1:index.html\n167.9.11.2:style.css"),
			forbiddenInfoList);
	}

	void testLoadForbiddenInfoListTwoRecordsEndOfLineInTheEndOfTheSource()
	{
		ForbiddenInfoList forbiddenInfoList;
		forbiddenInfoList.push_back(ForbiddenInfo("127.0.0.1", "index.html"));
		forbiddenInfoList.push_back(ForbiddenInfo("167.9.11.2", "style.css"));
		scenarioTestLoadForbiddenInfoListNoInfo(
			string("127.0.0.1:index.html\n167.9.11.2:style.css\n"),
			forbiddenInfoList);
	}

	void testLoadForbiddenInfoListFiveRecords()
	{
		ForbiddenInfoList forbiddenInfoList;
		forbiddenInfoList.push_back(ForbiddenInfo("127.0.0.1", "index.html"));
		forbiddenInfoList.push_back(ForbiddenInfo("167.9.11.2", "style.css"));
		forbiddenInfoList.push_back(ForbiddenInfo("4.9.11.2", "image.gif"));
		forbiddenInfoList.push_back(ForbiddenInfo("192.168.2.1", "text"));
		forbiddenInfoList.push_back(ForbiddenInfo("33.2.5.4", "ip.txt"));
		scenarioTestLoadForbiddenInfoListNoInfo(
			string("127.0.0.1:index.html\n"
				"167.9.11.2:style.css\n"
				"4.9.11.2:image.gif\n"
				"192.168.2.1:text\n"
				"33.2.5.4:ip.txt\n"),
			forbiddenInfoList);
	}

	void scenarioTestIsForbidden(const string &rawForbiddenInfo,
		const string &clientAddress, const string &objectToGetFileName,
		bool shouldBeForbidden)
	{
		auto_ptr<Server> server(new Server(40658));
		server->loadForbiddenInfoList(rawForbiddenInfo);

		CPPUNIT_ASSERT(server->isForbidden(clientAddress, objectToGetFileName) ==
			shouldBeForbidden);
	}

	void testIsForbiddenEmptyClientAddress()
	{
		scenarioTestIsForbidden(
			string("127.0.0.1:index.html\n167.9.11.2:style.css"),
			string(""), string(".html"),
			false);
	}

	void testIsForbiddenEmptyObjectToGetFileName()
	{
		scenarioTestIsForbidden(
			string("127.0.0.1:index.html\n167.9.11.2:style.css"),
			string("127.0.0.1"), string(""),
			false);
	}

	void testIsForbiddenIncludesOnlyAPartNotWholeFileName()
	{
		scenarioTestIsForbidden(
			string("127.0.0.1:.html\n167.9.11.2:style.css"),
			string("127.0.0.1"), string("test.html"),
			true);
	}

	void testIsForbiddenNotForbiddenClientAddress()
	{
		scenarioTestIsForbidden(
			string("127.0.0.1:index.html\n167.9.11.2:style.css"),
			string("127.0.0.2"), string("index.html"),
			false);
	}

	void testIsForbiddenNotForbiddenObjectToGetFileName()
	{
		scenarioTestIsForbidden(
			string("127.0.0.1:index.html\n167.9.11.2:style.css"),
			string("127.0.0.1"), string("index2.html"),
			false);
	}
};

CPPUNIT_TEST_SUITE_REGISTRATION(ServerTest);

//! Main.
/*!
	Runs all registered test suites.
*/
int main()
{
	TextUi::TestRunner runner;
	TestFactoryRegistry &registry = TestFactoryRegistry::getRegistry();
	runner.addTest(registry.makeTest());
	bool wasSuccessful = runner.run("", false);
	return !wasSuccessful;
}

/* End of file proxyhttp-tester.cc */
