/*
 * Encoding: UTF-8
 * File:     proxyhttp.cc
 * Project:  ISA - Simple HTTP proxy server
 * Date:     20.10.2007
 * Author:   Petr Zemek, xzemek02@stud.fit.vutbr.cz
 */

#include "proxyhttp.h"

namespace ProxyHTTP {

using std::cout;
using std::cerr;
using std::endl;
using std::string;
using std::vector;
using std::auto_ptr;
using std::exception;
using std::bad_alloc;
using std::ifstream;
using std::ostream;
using std::find_if;
using std::equal_to;
using std::getline;

////////////// Internal functions, types, constants and variables //////////////
namespace {

//! Case-insensitive comparatison of two chars.
/*!
	\param ch1 First character.
	\param ch2 Second character.
	\returns \a true if the \a ch1 is equal to \a ch2 (case-insensitive
	comparison), \a false otherwise.
*/
inline bool charsEqualCaseInsensitive(char ch1, char ch2)
{
	return std::toupper(static_cast<unsigned char>(ch1)) ==
		std::toupper(static_cast<unsigned char>(ch2));
}

//! Case-insensitive version of string.find().
/*!
	\param haystack String that will be searched for the \a needle occurence.
	\param needle String that is to be find in the \a haystack.
	\returns Index of the first occurence of \a needle in \a haystack,
	string::npos if \a needle is not found.
*/
string::size_type findCaseInsensitive(const string &haystack,
	const string &needle)
{
	string::const_iterator pos = search(haystack.begin(), haystack.end(),
		needle.begin(), needle.end(), charsEqualCaseInsensitive);

	return (pos == haystack.end()) ? string::npos : (pos - haystack.begin());
}

//! Splits \a source string by \a delimiter into \a parts vector.
/*!
	\param source String that will be split into substrings.
	\param delimiter Delimiter that will be used to separate strings.
	\param parts [out] Splitted strings will be saved into this vector.
*/
void explodeString(const string &source, char delimiter, vector<string> &parts)
{
	string::size_type partStart = 0;
	string::size_type partEnd = 0;

	do {
		partEnd = source.find(delimiter, partStart);

		string::size_type part_length = 0;
		if (partEnd == string::npos) {
			part_length = source.length() - partStart;
		} else {
			part_length = partEnd - partStart;
		}

		if (part_length > 0) {
			string part(source.substr(partStart, part_length));
			parts.push_back(part);
		}

		partStart = partEnd + 1;
	} while (partEnd != string::npos );
}

//! Prints program help to standard output.
/*!
*/
void printHelp()
{
	cout <<
		"Name: \n"
		"    proxyhttp - Simple HTTP proxy server\n"
		"Synopsis:\n"
		"    proxyhttp <local_port> <host>[:<port>]\n"
		"\n"
		"    local_port - port number (1 - 65535)\n"
		"    host - host address (DNS name)\n"
		"    port - host port number (1 - 65535, 80 by default)\n"
		"Description:\n"
		"    Simple HTTP proxy server (\"mediator\") that only receives HTTP\n"
		"    requests from clients, redirects them to the host server and\n"
		"    returns responses to clients.\n"
		"    If the request method is GET, client IP address\n"
		"    and the filename of the object (without directories) to be get\n"
		"    will be printed to standard output in the following format:\n"
		"        <client_ip>:<filename>\n"
		"    If there is a file called 'forbidden' in the server root directory,\n"
		"    every object filename will be searched for patterns listed\n"
		"    in the 'forbidden' file and if it matches, 404 (Not Found) response\n"
		"    will be sent to the client that requested the file and the following\n"
		"    output will be printed to standard error:\n"
		"        <client_ip>:<filename>\n"
		"    Every line in the 'forbidden' file must be in the following format:\n"
		"        <client_ip>:<filename>\n"
		"    Server can be stopped by sending SIGINT or SIGTERM signal to the\n"
		"    main proxyhttp server process.\n"
		"Examples:\n"
		"    proxyhttp 80 www.muni.cz:80\n"
		"    proxyhttp 8080 www.fit.vutbr.cz\n"
		"Author:\n"
		"    Petr Zemek, xzemek02@stud.fit.vutbr.cz, 2007"
		<< endl;
}

//! Prints selected \a errorMessage to standard error.
/*!
	\param errorMessage Error string that will be printed.
*/
inline void printErrorMessage(const string &errorMessage)
{
    cerr << "proxyhttp: " << errorMessage << endl;
}

//! Check whether the selected file exists.
/*!
	\param fileNameWithPath Path to the file.
	\returns \a true if the selected file is found, \a false otherwise.
*/
bool fileExists(const string &fileNameWithPath)
{
 	struct stat fileInfo;

	// Attempt to get the file attributes
	int intStat = stat(fileNameWithPath.c_str(), &fileInfo);
	if (intStat == 0) {
		// We were able to get the file attributes so the file obviously exists
		return true;
	}

	// We were not able to get the file attributes
	return false;
}

//! Reads all content from the selected file.
/*!
	\param fileNameWithPath Path to the file.
	\returns All content read from the selected file.

	If the selected file cannot be opened, RuntimeException will be thrown.
*/
string readAllFileContent(const string &fileNameWithPath)
{
	ifstream inputFile(fileNameWithPath.c_str());
	if (!inputFile) {
		throw RuntimeException(string("File `") + fileNameWithPath +
			string("` cannot be opened."));
	}

	string line;
	string inputFileContent;
	while (getline(inputFile, line))
		inputFileContent += line + "\n";

	return inputFileContent;
}

} // namespace (anonymous)

///////////////////////////// CommandLineArguments /////////////////////////////

//! Constructor.
/*!
	\param argc Number of arguments (like the one passed to main() function).
	\param argv Arguments "vector" (like the one passed to main() function).

	Creates and initalizes a CommandLineArguments object.
*/
CommandLineArguments::CommandLineArguments(int argc, const char **argv)
	: _localPort(0), _hostAddress(), _hostPort(DefaultHostPortNumber)
{
	assert(argv != 0);

	if (argc != MandatoryArgumentsCount) {
		throw CommandLineArgumentsException(string("Invalid arguments count."));
	}

	parsePortNumber(argv[LocalPortArgumentIndex], _localPort);
	parseHostData(argv[HostDataArgumentIndex]);
}

//! Destructor.
/*!
	Destroys a CommandLineArguments object.
*/
CommandLineArguments::~CommandLineArguments()
{
}

//! Check whether program help is requested to be shown.
/*!
	\param argc Number of arguments (like the one passed to main() function).
	\param argv Arguments "vector" (like the one passed to main() function).
	\returns \a true if program help is requested to be shown, \a false otherwise.
*/
bool CommandLineArguments::helpRequested(int argc, const char **argv)
{
	for (int i = 0; i < argc; ++i) {
		if (argv[i] == string("-h") || argv[i] == string("--help")) {
			return true;
		}
	}

	return false;
}

//! Parses port number from \a rawPortNumber and stores it into \a portNumber.
/*!
	\param rawPortNumber String with raw port number which will be parsed.
	\param portNumber [out] Into this variable the parsed port number will be stored.
*/
void CommandLineArguments::parsePortNumber(const string &rawPortNumber,
	unsigned int &portNumber)
{
	bool convertedOk = stringToNumber(rawPortNumber, portNumber);
	if (!convertedOk || !isValidPortNumber(portNumber)) {
		throw CommandLineArgumentsException(string("Invalid port number: ") +
			rawPortNumber + string("."));
	}
}

//! Parses host data (host address and optionaly a host port number).
/*!
	\param rawHostData String that contains host data.
*/
void CommandLineArguments::parseHostData(const string &rawHostData)
{
	assert(!rawHostData.empty());

	string::size_type firstColon = rawHostData.find_first_of(":");
	if (firstColon != string::npos) {
		string rawPortNumber(rawHostData, firstColon + 1, rawHostData.size());
		parsePortNumber(rawPortNumber, _hostPort);
	}

	_hostAddress = rawHostData.substr(0, firstColon);
	if (_hostAddress.empty()) {
		throw CommandLineArgumentsException(string("Host address is missing."));
	}
}

//! Validates selected port number.
/*!
	\param portNumber Port number to be validated.
	\returns \a true if the port number is valid, \a false otherwise.
*/
bool CommandLineArguments::isValidPortNumber(unsigned int portNumber)
{
	return portNumber >= MinimalPortNumber &&
		portNumber <= MaximalPortNumber;
}

const int CommandLineArguments::MandatoryArgumentsCount;
const int CommandLineArguments::LocalPortArgumentIndex;
const int CommandLineArguments::HostDataArgumentIndex;
const unsigned int CommandLineArguments::DefaultHostPortNumber;
const unsigned int CommandLineArguments::MinimalPortNumber;
const unsigned int CommandLineArguments::MaximalPortNumber;

///////////////////////////////// HttpMessage //////////////////////////////////

//! Default constructor.
/*!
	\param message Raw HTTP message.

	Creates and initalizes a HttpMessage object.
*/
HttpMessage::HttpMessage(string message)
	: _message(message)
{
}

//! Destructor.
/*!
	Destroys a HttpMessage object.
*/
HttpMessage::~HttpMessage()
{
}

//! \returns stored raw HTTP message.
string HttpMessage::rawMessage() const
{
	return _message;
}

//! \returns raw HTTP message body (without headers).
string HttpMessage::rawMessageBody() const
{
	string::size_type messageBodyStartIndex = _message.find(HeaderAndBodyDelimiter);
 	return _message.substr(messageBodyStartIndex +
 		string(HeaderAndBodyDelimiter).size());
}

//! \returns HTTP version that the HTTP message contains.
string HttpMessage::httpVersion() const
{
	if (matchesHttpVersion(_message, HttpVersion10Marker)) {
		return HttpVersion10Marker;
	} else if (matchesHttpVersion(_message, HttpVersion11Marker)) {
		return HttpVersion11Marker;
	}

	return string("");
}

//! \returns value of the header named \a headerName.
/*!
	\param headerName Name of the header which value will be returned.
*/
string HttpMessage::getHeaderValue(const string &headerName) const
{
	if (headerName.empty()) {
		return string("");
	}

	string::size_type headerNameStart = findCaseInsensitive(_message,
		headerName + string(HeaderValueDelimiter));
	if (headerNameStart != string::npos) {
		const string::size_type headerValueStartIndex = headerNameStart +
			string(headerName + string(HeaderValueDelimiter)).size();
		const string messageStartingWithHeaderValue = _message.substr(
			headerValueStartIndex);
		return messageStartingWithHeaderValue.substr(0,
			messageStartingWithHeaderValue.find(HeaderEndOfLineMarker));
	}

	return string("");
}

//! Changes selected \a headerName value to \a newHeaderValue.
/*!
	\param headerName Name of the header.
	\param newHeaderValue New header value that will be set.
	\returns \a true if there was a change, \a false otherwise.
*/
bool HttpMessage::changeHeaderValue(const string &headerName,
	const string &newHeaderValue)
{
	if (headerName.empty() || newHeaderValue.empty()) {
		return false;
	}

	string oldHeaderValue = getHeaderValue(headerName);
	if (oldHeaderValue.empty()) {
		return false;
	}

	string::size_type headerValueToReplaceStartIndex = findCaseInsensitive(
		_message, headerName + HeaderValueDelimiter) + headerName.size() +
		string(HeaderValueDelimiter).size();
	string::size_type headerValueToReplaceEndIndex = headerValueToReplaceStartIndex +
		oldHeaderValue.size();

	// Finally, do the replacement
	_message = _message.substr(0, headerValueToReplaceStartIndex) +
		newHeaderValue + _message.substr(headerValueToReplaceEndIndex);
	return true;
}

//! \returns \a true if the connection has to be persistent, \a false otherwise
/*!
	"Connection has to be persistent" means that it either is a HTTP/1.1 message
	and it doesn't have "Connection: close" header, or it is a HTTP/1.0 message
	with "Connection: keep-alive" header.
*/
bool HttpMessage::connectionHasToBePersistent() const
{
	bool isHttp10Message = httpVersion() == HttpVersion10Marker;
	bool isHttp11Message = httpVersion() == HttpVersion11Marker;
	// I'm doing case-insensitive comparation because some servers/browsers
	// send "close", but others send "Close" etc...
	bool hasConnectionCloseHeader = getHeaderValue("Connection") == string("close") ||
		getHeaderValue("Connection") == string("Close");;
	bool hasConnectionKeepAliveHeader = getHeaderValue("Connection") == string("keep-alive") ||
		getHeaderValue("Connection") == string("Keep-Alive");
	return isHttp10Message && hasConnectionKeepAliveHeader ||
		isHttp11Message && !hasConnectionCloseHeader;
}

//! \returns \a true if the response has "Content-Length" header present, \a false otherwise.
bool HttpMessage::hasContentLengthHeader() const
{
	return !getHeaderValue(ContentLengthHeaderName).empty();
}

//! \returns \a true if the response is complete by "Content-Length: x", \a false otherwise.
bool HttpMessage::isCompleteByContentLength() const
{
	string rawContentLength = getHeaderValue(ContentLengthHeaderName);
	if (rawContentLength.empty()) {
		return false;
	}

	string::size_type contentLength = 0;
	bool convertedOk = stringToNumber(rawContentLength, contentLength);
	if (!convertedOk) {
		return false;
	}

	if (rawMessageBody().size() < contentLength) {
		return false;
	}

	return true;
}

const char *HttpMessage::HttpVersion10Marker = "HTTP/1.0";
const char *HttpMessage::HttpVersion11Marker = "HTTP/1.1";
const char *HttpMessage::HeaderValueDelimiter = ": ";
const char *HttpMessage::HeaderAndBodyDelimiter = "\r\n\r\n";
const char *HttpMessage::HeaderEndOfLineMarker = "\r\n";
const char *HttpMessage::ContentLengthHeaderName = "Content-Length";

///////////////////////////////// HttpRequest //////////////////////////////////

//! Default constructor.
/*!
	\param requestMessag Raw HTTP request message.

	Creates and initalizes a HttpRequest object.
*/
HttpRequest::HttpRequest(string requestMessage)
	: HttpMessage(requestMessage)
{
}

//! Destructor.
/*!
	Destroys a HttpRequest object.
*/
HttpRequest::~HttpRequest()
{
}

//! \returns type of the HTTP request (GET, POST, ...).
string HttpRequest::httpRequestType() const
{
	string::size_type firstSpaceIndex = _message.find(" ");
	return _message.substr(0, firstSpaceIndex);
}

//! \returns Object to be get via GET without directory path.
/*!
	If the requested file (path) ends with '/' like in this request:
	"GET / HTTP/1.1", HttpRequest::DefaultObjectFileName will be returned.
	If the object filename is not found, empty string will be returned.
*/
string HttpRequest::objectToGetfileName() const
{
	if (httpRequestType() != HttpRequestTypeGet) {
		return string("");
	}

	string::size_type firstSpaceIndex = _message.find(" ");
	if (firstSpaceIndex == string::npos) {
		return string("");
	}

	string::size_type secondSpaceIndex = _message.substr(firstSpaceIndex + 1).find(" ");
	if (secondSpaceIndex == string::npos) {
		return string("");
	}
	secondSpaceIndex += firstSpaceIndex + 1;

	string fullPathWithObjectFileName = _message.substr(firstSpaceIndex + 1,
		secondSpaceIndex - firstSpaceIndex - 1);
	// Remove URL arguments after the first '?' (searchpart)
	string::size_type argumentsStartIndex = fullPathWithObjectFileName.find("?");
	if (argumentsStartIndex != string::npos) {
		fullPathWithObjectFileName = fullPathWithObjectFileName.substr(
			0, argumentsStartIndex);
	}

	// Check whether the path does not contain subdirectories
	string::size_type lastSlashIndex = fullPathWithObjectFileName.find_last_of("/");
	if (lastSlashIndex == string::npos) {
		// There is no slash, so fullPathWithObjectFileName
		// must contain filename
		return fullPathWithObjectFileName;
	}

	return (lastSlashIndex == (fullPathWithObjectFileName.size() - 1)) ?
		string(DefaultObjectFileName) : fullPathWithObjectFileName.substr(lastSlashIndex + 1);
}

//! \returns \a true if the request is complete, \a false otherwise.
/*!
	"Complete" request means that it ends with "\r\n\r\n".
*/
bool HttpRequest::isComplete() const
{
	if (hasContentLengthHeader()) {
		return isCompleteByContentLength();
	}

	// Requests without body
 	return _message.find(HeaderAndBodyDelimiter) != string::npos;
}

//! Check whether the passed \a httpVersionMarker matches the one in \a requestMessage.
/*!
	\param requestMessage Raw HTTP request message.
	\param httpVersionMarker HTTP version (like "HTTP/1.1" for HTTP 1.1 version).
	\returns \a true if the \a httpVersionMarker was found in the \a requestMessage,
	\a false otherwise.
*/
bool HttpRequest::matchesHttpVersion(const string &requestMessage,
	const string &httpVersionMarker) const
{
	const string::size_type firstEndOfLineIndex = requestMessage.find(HeaderEndOfLineMarker);
	const string firstLine = requestMessage.substr(0, firstEndOfLineIndex +
		(firstEndOfLineIndex == string::npos ? 0 : string(HeaderEndOfLineMarker).size()));
	return firstLine.find(httpVersionMarker + string(HeaderEndOfLineMarker)) != string::npos;
}

//! \returns a new (single) HttpRequest from \a rawData.
/*!
	\param rawData Input (output) data containing 0-* HTTP requests.

	Returned HttpRequest text is removed from the \a rawData string, so
	you can use this function in a loop for getting all of the requests
	from \a rawData.
*/
HttpRequest HttpRequest::getSingleHttpRequest(string &rawData)
{
	string::size_type endOfRequestIndex = rawData.find(
		string(HttpRequest::HeaderAndBodyDelimiter));
	if (endOfRequestIndex != string::npos) {
		string::size_type newRequestStartIndex = endOfRequestIndex +
			string(HeaderAndBodyDelimiter).size();
		HttpRequest httpRequest(rawData.substr(0, newRequestStartIndex));
		rawData = rawData.substr(newRequestStartIndex);

		// Some requests might have a message body
		if (httpRequest.hasContentLengthHeader()) {
			// Get content length
			string::size_type contentLength = 0;
			bool convertedOk = stringToNumber(httpRequest.getHeaderValue(
				string(ContentLengthHeaderName)), contentLength);
			if (!convertedOk) {
				// Invalid request
				return HttpRequest(string(""));
			}

			if (rawData.size() < contentLength) {
				// Request is not complete, so restore the previous rawData
				// content
				rawData = httpRequest.rawMessage() + rawData;
				return HttpRequest(string(""));
			}

			// Append message body to the request
			httpRequest = HttpRequest(httpRequest.rawMessage() +
				rawData.substr(0, contentLength));
			rawData = rawData.substr(contentLength);
		}
		return httpRequest;
	}

	return HttpRequest(string(""));
}

const char *HttpRequest::HttpRequestTypeGet = "GET";
const char *HttpRequest::DefaultObjectFileName = "index.html";
const char *HttpRequest::HostHeaderName = "Host";

///////////////////////////////// HttpResponse /////////////////////////////////

//! Default constructor.
/*!
	\param responseMessage Raw HTTP response message.

	Creates and initalizes a HttpRequest object.
*/
HttpResponse::HttpResponse(string responseMessage)
	: HttpMessage(responseMessage)
{
}

//! Destructor.
/*!
	Destroys a HttpResponse object.
*/
HttpResponse::~HttpResponse()
{
}

//! Check whether the passed \a httpVersionMarker matches the one in \a responseMessage.
/*!
	\param responseMessage Raw HTTP request message.
	\param httpVersionMarker HTTP version (like "HTTP/1.1" for HTTP 1.1 version).
	\returns \a true if the \a httpVersionMarker was found in the \a requestMessage,
	\a false otherwise.
*/
bool HttpResponse::matchesHttpVersion(const string &responseMessage,
	const string &httpVersionMarker) const
{
	return responseMessage.substr(0, httpVersionMarker.size()) == httpVersionMarker;
}

//! \returns \a true if the response is complete, \a false otherwise.
/*!
	"Complete" response means:
	- there is a "Content-Length: x" header and the response body has "x"
	  or more bytes.
	- there is a "Transfer-Encoding: chunked" header and the response body
	  ends with "\r\n0\r\n\r\n".
	- there is no response body, but there are some headers.
*/
bool HttpResponse::isComplete() const
{
	if (hasTransferEncodingChunkedHeader()) {
		return isCompleteByTransferEncodingChunked();
	} else if (hasContentLengthHeader()) {
		return isCompleteByContentLength();
	}

 	// There is neither "Content-Length: x" nor "Transfer-Encoding: chunked"
 	// header so check whether there already is a body/header delimiter
 	return _message.find(HeaderAndBodyDelimiter) != string::npos;
}

//! \returns \a true if the response is complete by "Transfer-Encoding: chunked", \a false otherwise.
bool HttpResponse::isCompleteByTransferEncodingChunked() const
{
	if (!hasTransferEncodingChunkedHeader()) {
		return false;
	}

	// There must be the "Transfer-Encoding: chunked" end marker
	// at the end of the message body
	string foundEndMarker = rawMessageBody().substr(rawMessageBody().size() -
		string(TransferEncodingChunkedEndMarker).size());
	return foundEndMarker == TransferEncodingChunkedEndMarker;
}

//! \returns \a true if the response has "Content-Length" header present
bool HttpResponse::hasTransferEncodingChunkedHeader() const
{
	return getHeaderValue(TransferEncodingHeaderName) == TransferEncodingChunkedValue;
}

//! \returns HTTP response code or 0, if not present or there is some error.
unsigned int HttpResponse::responseCode() const
{
	if (_message.size() <= ResponseCodeStartIndex) {
		return 0;
	}

	string rawResponseCode = _message.substr(ResponseCodeStartIndex,
		ResponseCodeLength);
	unsigned int responseCode = 0;
	// Return value is ignored on purpose because this function
	// returns 0 if an error occurs
	stringToNumber(rawResponseCode, responseCode);
	return responseCode;
}

//! Creates and returns a HTTP 404 Not Found response.
/*!
	\param httpVersion HTTP version of the created response.
	\returns 404 Not Found response.
*/
HttpResponse HttpResponse::createHttpNotFoundResponse(const string &httpVersion)
{
	// Create message body
	string messageBody("<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">"
		"<HTML><HEAD><TITLE>404 Not Found</TITLE></HEAD><BODY>"
		"<H1>Not Found</H1>"
		"The requested URL was not found on this server.<P>"
		"<HR></BODY></HTML>");
	// Create headers
	string headers;
	headers += string("Content-Type: text/html; charset=iso-8859-1") +
		string(HeaderEndOfLineMarker);
	headers += string("Content-Length:") + toString(messageBody.size()) +
		string(HeaderEndOfLineMarker);
	// Finally, create the 404 response
	return HttpResponse(httpVersion + string(" 404 Not Found") +
		string(HeaderEndOfLineMarker) +
		headers +
		string(HeaderEndOfLineMarker) +
		messageBody);
}

const char *HttpResponse::TransferEncodingHeaderName = "Transfer-Encoding";
const char *HttpResponse::TransferEncodingChunkedValue = "chunked";
const char *HttpResponse::TransferEncodingChunkedEndMarker = "\r\n0\r\n\r\n";
const unsigned int HttpResponse::ResponseCodeStartIndex;
const unsigned int HttpResponse::ResponseCodeLength;

////////////////////////////////// Connection //////////////////////////////////

//! Default constructor.
/*!
	Creates and initalizes a Connection object.
*/
Connection::Connection() {}

//! Destructor.
/*!
	Destroys a Connection object.
*/
Connection::~Connection() {}

/////////////////////////////// ConnectionToHost ///////////////////////////////

//! Constructor.
/*!
	\param hostAddress DNS address of the remote host.
	\param hostPort Port number of the remote host.

	Creates and initalizes a ConnectionToHost object by creating a TCP
	connection with the selected remote host.
*/
ConnectionToHost::ConnectionToHost(string hostAddress, unsigned int hostPort)
	: _hostAddress(hostAddress), _hostPort(hostPort),
	  _hostSocket(0), _hostSocketInfo()
{
	connectToHost();
}

//! Destructor.
/*!
	Destroys a ConnectionToHost object.
*/
ConnectionToHost::~ConnectionToHost()
{
	closeConnection();
}

//! \returns Host address (DNS name).
string ConnectionToHost::hostAddress() const
{
	return _hostAddress;
}

//! \returns Host port number.
unsigned int ConnectionToHost::hostPort() const
{
	return _hostPort;
}

//! Creates a TCP connection with the remote host.
/*!
	If the connection cannot be established, NetworkException will be thrown.
*/
void ConnectionToHost::connectToHost()
{
	// Transform hostAddress into "usable" form
	errno = 0;
	hostent *hostInfo = gethostbyname(_hostAddress.c_str());
	if (!hostInfo) {
		int errnoBackup = errno;
		string reason = errno > 0 ? string(strerror(errnoBackup)) :
			string("Invalid host address");
		throw NetworkException(string("Host information retrieval failed. Reason: ") +
			 reason + string("."));
	}

	// Create socket
	errno = 0;
	if ((_hostSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		int errnoBackup = errno;
		throw NetworkException(string("Socket creation failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}

	// Fill sockaddr_in structure with protocol family, host port number
	// and host IP address
	_hostSocketInfo.sin_family = AF_INET;
	_hostSocketInfo.sin_port = htons(_hostPort);
	memcpy(&_hostSocketInfo.sin_addr, hostInfo->h_addr, hostInfo->h_length);

	// Establish connection
	// (reinterpret_cast<> is necessary here...)
	errno = 0;
	if (connect(_hostSocket, reinterpret_cast<sockaddr *>(&_hostSocketInfo), sizeof(_hostSocketInfo)) == -1) {
		int errnoBackup = errno;
		throw NetworkException(string("Connection to the selected server failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}
}

//! Reconnects with the remote host.
void ConnectionToHost::reconnectToHost()
{
	closeConnection();
	connectToHost();
}

//! Closed TCP connection with the remote host.
void ConnectionToHost::closeConnection()
{
	close(_hostSocket);
}

//! Sends HTTP \a request to the remote host.
/*!
	\param request HTTP request that will be send to the remote host.
*/
void ConnectionToHost::sendRequest(const HttpRequest &request)
{
	const char *rawMessage = request.rawMessage().c_str();
	const size_t rawMessageSize = request.rawMessage().size();
	if (send(_hostSocket, rawMessage, rawMessageSize, 0) == -1) {
		// Try to reconnect and resend the request
		reconnectToHost();
		errno = 0;
		if (send(_hostSocket, rawMessage, rawMessageSize, 0) == -1) {
			int errnoBackup = errno;
			throw NetworkException(string("Request transmission failed. Reason: ") +
				string(strerror(errnoBackup)) + string("."));
		}
	}
}

//! \returns Received response from the remote host.
/*!
	If the response cannot be received it returns empty HTTP response.
*/
HttpResponse ConnectionToHost::receiveResponse()
{
	string responseMessage;
	int receivedBytes = 0;
	while ((receivedBytes = recv(_hostSocket, buffer, BufferSize - 1, 0)) > 0) {
 		buffer[receivedBytes] = '\0';
		responseMessage += string(buffer, receivedBytes);

		HttpResponse httpResponse(responseMessage);
		if (httpResponse.connectionHasToBePersistent() && httpResponse.isComplete()) {
			return httpResponse;
		}
	}

	return HttpResponse(responseMessage);
}

///////////////////////////// ConnectionFromClient /////////////////////////////

//! Constructor.
/*!
	\param clientSocket Created and connected socket.
	\param clientSocketInfo Initialized sockaddr_in structure.

	Creates and initalizes a ConnectionFromClient object.
*/
ConnectionFromClient::ConnectionFromClient(int clientSocket,
	sockaddr_in clientSocketInfo)
	: _clientSocket(clientSocket), _clientSocketInfo(clientSocketInfo),
	  buffer(), rawData("")
{
}

//! Destructor.
/*!
	Destroys a ConnectionFromClient object. Connection will be closed if it
	is opened.
*/
ConnectionFromClient::~ConnectionFromClient()
{
	close(_clientSocket);
}

//! \returns Client (IPv4) address.
string ConnectionFromClient::clientAddress() const
{
	return inet_ntoa(_clientSocketInfo.sin_addr);
}

//! \returns Client port number.
unsigned int ConnectionFromClient::clientPort() const
{
	return ntohs(_clientSocketInfo.sin_port);
}

//! Closed TCP connection with the client.
void ConnectionFromClient::closeConnection()
{
	close(_clientSocket);
}

//! Sends selected HTTP \a response to the client.
/*!
	\param response HTTP response that will be sent to the client.

	If the response failed to be sent, NetworkException will be thrown.
*/
void ConnectionFromClient::sendResponse(const HttpResponse &response)
{
	const char *rawMessage = response.rawMessage().c_str();
	const size_t rawMessageSize = response.rawMessage().size();
	errno = 0;
	if (send(_clientSocket, rawMessage, rawMessageSize, 0) == -1) {
		int errnoBackup = errno;
		throw NetworkException(string("Request transmission failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}
}

//! Receives a new HTTP request from the client.
/*!
	\returns Received HTTP request.

	If there is a problem with receiving the request, empty request will
	be returned.
*/
HttpRequest ConnectionFromClient::receiveRequest()
{
	// Check whether there already is a complete request in the buffer
	HttpRequest request = HttpRequest::getSingleHttpRequest(rawData);
	if (request.isComplete()) {
		return request;
	}

	int receivedBytes = 0;
	while ((receivedBytes = recv(_clientSocket, buffer, BufferSize - 1, 0)) > 0) {
 		buffer[receivedBytes] = '\0';
		rawData += string(buffer, receivedBytes);

		request = HttpRequest::getSingleHttpRequest(rawData);
		if (request.isComplete()) {
			return request;
		}
	}

	return HttpRequest(string(""));
}

const int ConnectionFromClient::BufferSize;

//////////////////////////////////// Server ////////////////////////////////////

//! Constructor.
/*!
	\param localPort Local port that the server will be listening on.

	Creates and initializes a Server object that will be listening on
	the selected port.
*/
Server::Server(unsigned int localPort)
	: _localPort(localPort), _localSocket(0), _localSocketInfo(),
	  blockMaskCurrent(), blockMaskOld(), sigchldHandler(), sigchldHandlerOld(),
	  _forbiddenInfoList()
{
	// Create socket
	errno = 0;
	if ((_localSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		int errnoBackup = errno;
		throw NetworkException(string("Socket creation failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}

	// Fill sockaddr_in structure with protocol family, port number
	// and IP address (this is set to listen on any interface)
	_localSocketInfo.sin_family = AF_INET;
	_localSocketInfo.sin_port = htons(_localPort);
	_localSocketInfo.sin_addr.s_addr = htonl(INADDR_ANY);

	// Set SO_REUSEADDR on the socket to true (1)
	// This prevents errors when reusing the port number (for example when
	// restarting this the server on the same port in a very short time interval)
	int reuseAddrOption = 1;
	errno = 0;
	if (setsockopt(_localSocket, SOL_SOCKET, SO_REUSEADDR, &reuseAddrOption, sizeof(reuseAddrOption)) == -1) {
		int errnoBackup = errno;
		throw NetworkException(string("Setting port number reuse option failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}

	// Bind the "name" (address) to the socket
	// (reinterpret_cast<> is necessary here...)
	errno = 0;
	if (bind(_localSocket, reinterpret_cast<sockaddr *>(&_localSocketInfo), sizeof(_localSocketInfo)) == -1) {
		int errnoBackup = errno;
		close(_localSocket);
		throw NetworkException(string("Name bind to the socket failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}

	// Create a client connection queue
	errno = 0;
	if (listen(_localSocket, MaxConnectionCount) == -1) {
		int errnoBackup = errno;
		close(_localSocket);
		throw NetworkException(string("Client connection queue creation failed. Reason: ") +
			string(strerror(errnoBackup)) + string("."));
	}

	setSignalHandling();
}

//! Destructor.
/*!
	Destroys a Server object and stops listening (if it's listening).
*/
Server::~Server()
{
	close(_localSocket);
	restoreSignalHandling();
}

//! Loads a list of forbidden files to get from \a rawForbiddenInfo string.
/*!
	\param rawForbiddenInfo Raw forbiden files info which will be parsed.

	\a rawForbiddenInf must be in the following format:
		client_address:filename1\n
		client_address:filename2\n
		...
*/
void Server::loadForbiddenInfoList(const string &rawForbiddenInfo)
{
	vector<string> lines;
	explodeString(rawForbiddenInfo, ForbiddenInfoRecordsDelimiter, lines);

	for (vector<string>::iterator lineIter = lines.begin();
		 lineIter != lines.end(); ++lineIter) {
		string currentLine = *lineIter;
		string::size_type currentIndex = 0;

		string address;
		while (currentIndex < currentLine.size() &&
			currentLine[currentIndex] != ForbiddenInfoRecordPartsDelimiter) {
			address += currentLine[currentIndex];
			++currentIndex;
		}
		++currentIndex; // Skip ':'

		string objectFileName;
		while (currentIndex < currentLine.size()) {
			objectFileName += currentLine[currentIndex];
			++currentIndex;
		}

		if (address.empty() || objectFileName.empty()) {
			throw RuntimeException(string(
				"Data in the 'forbidden' file have invalid format."));
		}

		_forbiddenInfoList.push_back(ForbiddenInfo(address, objectFileName));
	}
}

//! Checks whether the selected combination (address:filename) is forbidden.
/*!
	\param clientAddress Client address.
	\param objectFileName Filename of the forbidden object.
	\returns \a true if the selected combination is forbidden, \a false otherwise.
*/
bool Server::isForbidden(const string &clientAddress, const string &objectFileName)
{
	for (ForbiddenInfoList::iterator iter = _forbiddenInfoList.begin();
		 iter != _forbiddenInfoList.end(); ++iter) {
		if (iter->first == clientAddress &&
			(objectFileName.find(iter->second) != string::npos)) {
			return true;
		}
	}

	return false;
}

//! Creates a connection to the selected host.
/*!
	\param hostAddress Address of the host (DNS name).
	\param hostPort Port that the host is listening on.
	\returns Created connection to the selected host.
*/
auto_ptr<ConnectionToHost> Server::connectToHost(string hostAddress,
	unsigned int hostPort)
{
	return auto_ptr<ConnectionToHost>(new ConnectionToHost(hostAddress, hostPort));
}

//! Accepts a connection from a client.
/*!
	\returns Accepted and "created" connection from a client.

	If the acceptance fails, NetworkException will be thrown.
*/
auto_ptr<ConnectionFromClient> Server::acceptConnection()
{
	sockaddr_in clientSocketInfo;
	socklen_t clientSocketInfoSize = sizeof(clientSocketInfo);

	// (reinterpret_cast<> is necessary here...)
	errno = 0;
	int clientSocket = accept(_localSocket,
		reinterpret_cast<sockaddr *>(&clientSocketInfo), &clientSocketInfoSize);
	if (clientSocket == -1) {
		int errnoBackup = errno;
		throw NetworkException(string("Client connection acceptance failed (") +
			inet_ntoa(clientSocketInfo.sin_addr) + "). Reason: " +
			string(strerror(errnoBackup)) + string("."));
	}

	return auto_ptr<ConnectionFromClient>(new ConnectionFromClient(
		clientSocket, clientSocketInfo));
}

//! Completely processes the selected connection from a client.
/*!
	\param connectionFromClient Established connection from a client.
	\param hostAddress Host address of the remote host.
	\param hostPort Port number of the remote host.

	Selected connection is processed in a new process created using fork()
	function, so in parent process this function will return immediately
	and child process will exit (using exit() function) after the connection
	is processed. Processing means that the server receives requests from
	the client, redirects them to the remote host and received responses
	redirects to the client.
*/
void Server::processConnectionFromClient(
	std::auto_ptr<ConnectionFromClient> connectionFromClient,
	const std::string &hostAddress, unsigned int hostPort)
{
	// Fork a new process to handle that connection (concurrent server behavior)
	errno = 0;
	pid_t pid = fork();
	if (pid < 0) {
		int errnoBackup = errno;
		throw RuntimeException(string("Creation of a new server process "
			"failed (fork). Reason: ") + string(strerror(errnoBackup)) + string("."));
	} else if (pid == 0) {
		// Child process, handle the accepted client

		// Create connection to the remote host
		std::auto_ptr<ConnectionToHost> connectionToHost(
			connectToHost(hostAddress, hostPort));

		// These variables are defined here because they're needed in the
		// following do...while loop condition
		HttpRequest request;
		HttpResponse response;
		do {
			request = connectionFromClient->receiveRequest();
			if (!request.isComplete()) {
				break;
			}

			// Proccess the received request to see whether the client
			// is allowed to send this kind of request
			bool allowRequest = processHttpRequest(connectionToHost.get(),
				connectionFromClient.get(), request);
			if (allowRequest) {
				while (true) {
					// Request was processed successfully and can be send
					// to the client
					connectionToHost->sendRequest(request);

					// Receive response from the remote host
					// and "redirect" it to the client
					response = connectionToHost->receiveResponse();
					if (!response.isComplete()) {
						// There was an error in the sendRequest call,
						// possibly the TCP connection was dropped,
						// so try to resend the request
						continue;
					}
					connectionFromClient->sendResponse(response);
					break;
				}
			} else {
				// Request is denied, so send 'Not found' response
				// to the client
				response = HttpResponse::createHttpNotFoundResponse(
					request.httpVersion());
				connectionFromClient->sendResponse(response);
			}
		} while (request.connectionHasToBePersistent() &&
			response.connectionHasToBePersistent());

		// Child exists - close all connections
		connectionFromClient->closeConnection();
		connectionToHost->closeConnection();
		exit(EXIT_SUCCESS);
	}
}

//! Processes selected HTTP request.
/*!
	\param connectionToHost Initialized connection to a remote host.
	\param connectionFromClient Initialized connection from a client.
	\param originalHttpRequest Original HTTP request that will be processed.
	\returns \a true if the request is allowed, \a false otherwise (that
	means that the combination of client address and filename is forbidden).

	If the request type is GET, client address and the object to be get file
	name will be printed on the standard output. If the comination of client
	address and filename is forbidden, client address and the object to be get
	filename will be printed on the standard error instead.
*/
bool Server::processHttpRequest(ConnectionToHost *connectionToHost,
	ConnectionFromClient *connectionFromClient, HttpRequest &originalHttpRequest)
{
	bool returnValue = true;
	if (originalHttpRequest.httpRequestType() == HttpRequest::HttpRequestTypeGet) {
		string clientAddress = connectionFromClient->clientAddress();
		string objectToGetFileName = originalHttpRequest.objectToGetfileName();

		if (isForbidden(clientAddress, objectToGetFileName)) {
			// Forbidden info
			printRequestInfo(cerr, clientAddress, objectToGetFileName);
			returnValue = false;
		} else {
			// Standard info
			printRequestInfo(cout, clientAddress, objectToGetFileName);
		}
	}

	// Change "Host: " header value
	// Well, I should first check whether it is a HTTP/1.1 request because
	// HTTP/1.0 requests shouldn't have this header, but for example
	// `wget` sends this header even it sends HTTP/1.0 requests...
	originalHttpRequest.changeHeaderValue(string(HttpRequest::HostHeaderName),
		connectionToHost->hostAddress());

	return returnValue;
}

//! Sets signal handling.
/*!
	All signals except SIGINT, SIGTERM, SIGSTOP, SIGKILL and SIGCHLD
	will be blocked.

	SIGCHLD signal is ignored to avoid zombie processes when running
	a concurrent server.
*/
void Server::setSignalHandling()
{
	// Establish signal handler for SIGCHLD to avoid zombie processes
	sigemptyset(&sigchldHandler.sa_mask);
	sigchldHandler.sa_handler = SIG_IGN;
	sigchldHandler.sa_flags = 0;
	sigaction(SIGCHLD, &sigchldHandler, &sigchldHandlerOld);

	// Set up blocking mask for all processes
	sigfillset(&blockMaskCurrent);
	sigdelset(&blockMaskCurrent, SIGCHLD);
	sigdelset(&blockMaskCurrent, SIGINT);
	sigdelset(&blockMaskCurrent, SIGTERM);
	sigprocmask(SIG_BLOCK, &blockMaskCurrent, &blockMaskOld);
}

//! Restores signal handling.
void Server::restoreSignalHandling()
{
	sigaction(SIGCHLD, &sigchldHandlerOld, 0);
	sigprocmask(SIG_BLOCK, &blockMaskOld, 0);
}

//! Prints selected combination (address, object filename) to the selected stream.
/*!
	\param outputStream To this stream the info will be printed.
	\param clientAddress Address of a client.
	\param objectFileName Object filename.

	Printed info will have the following format:
		clientAddress:objectFileName
*/
void Server::printRequestInfo(ostream &outputStream,
	const string &clientAddress, const string &objectFileName)
{
	outputStream << clientAddress << ":" << objectFileName << endl;
}

const int Server::MaxConnectionCount;
const char Server::ForbiddenInfoRecordsDelimiter;
const char Server::ForbiddenInfoRecordPartsDelimiter;
const char *Server::ForbiddenInfoFileName = "forbidden";

} // namespace ProxyHTTP

///////////////////////////////////// Main /////////////////////////////////////

#ifndef PROXYHTTP_TESTER
//! Main.
/*!
	This is the program main entry point.
*/
int main(int argc, const char **argv)
{
	using namespace ProxyHTTP;

	try {
		if (CommandLineArguments::helpRequested(argc, argv)) {
			printHelp();
			return EXIT_SUCCESS;
		}

		// Parse command line arguments
		CommandLineArguments args(argc, argv);

		// Create a proxy server and initialize forbiden files info list
		auto_ptr<Server> server(new Server(args.localPort()));
		if (fileExists(Server::getForbiddenInfoFileName())) {
			server->loadForbiddenInfoList(readAllFileContent(
				Server::getForbiddenInfoFileName()));
		}

		// Test connection to the remote host
		std::auto_ptr<ConnectionToHost> testingConnectionToHost(
			new ConnectionToHost(args.hostAddress(), args.hostPort()));
		testingConnectionToHost->closeConnection();

		// Accept and handle connections from clients in an endless loop
		while (true) {
			try {
				auto_ptr<ConnectionFromClient> connectionFromClient =
					server->acceptConnection();

				server->processConnectionFromClient(connectionFromClient,
					args.hostAddress(), args.hostPort());
			}
			catch (const BaseException &e) {
				printErrorMessage(e.what());
			}
		}
	}
	catch (const CommandLineArgumentsException &e) {
		printErrorMessage(e.what());
		printErrorMessage("Try `proxyhttp --help' for more information.");
		return EXIT_FAILURE;
	}
	catch (const BaseException &e) {
		printErrorMessage(e.what());
		return EXIT_FAILURE;
	}
	catch (const bad_alloc &e) {
		printErrorMessage("Innsufficient memory.");
		return EXIT_FAILURE;
	}
	catch (const exception &e) {
		printErrorMessage(e.what());
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
#endif

/* End of file proxyhttp.cc */
