/*
 * Encoding: UTF-8
 * File:     proxyhttp.h
 * Project:  ISA - Simple HTTP proxy server
 * Date:     20.10.2007
 * Author:   Petr Zemek, xzemek02@stud.fit.vutbr.cz
 */

#ifndef PROXYHTTP_H
#define PROXYHTTP_H

// Debugging enabled (1) / disabled (0)
#define PROXYHTTP_DEBUG 0

#if !PROXYHTTP_DEBUG
	// Disables assert() macros
	#ifndef NDEBUG
		#define NDEBUG
	#endif
#endif

#include <iostream>
#include <fstream>
#include <exception>
#include <stdexcept>
#include <algorithm>
#include <functional>
#include <memory>
#include <string>
#include <vector>
#include <sstream>
#include <cctype>
#include <cassert>
#include <cerrno>
#include <unistd.h>
#include <signal.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

namespace ProxyHTTP {

//! Converts selected string \a str into a \a number.
/*!
	\param str String that will be converted into \a number.
	\param number Into this parameter the result number will be saved.
	\param format Number format (e.g. std::dec, std::hex).
 	\return Returns true if the conversion was successful, false otherwise.

	If the conversion fails, number value will be left unchanged.
*/
template<typename N> inline bool stringToNumber(const std::string & str,
	N & number, std::ios_base & (* format)(std::ios_base &) = std::dec)
{
	return !(std::istringstream(str) >> format >> number).fail();
}

//! Converts selected value into string.
/*!
	\param value Value that will be converted into string.
	\returns Converted \a value into string.
*/
template<typename T> inline std::string toString(const T & value)
{
	std::ostringstream out;
	out << value;
	return out.str();
}

//! Base exception for all HttpProxy exceptions.
/*!
	All HttpProxy exceptions are derived from this class.
*/
class BaseException: public std::exception {
public:
	explicit			BaseException(const std::string &e) throw(): s(e) {}
	virtual				~BaseException() throw() {}
	virtual const char	*what() throw() { return s.c_str(); }
	virtual const char	*what() const throw() { return s.c_str(); }

private:
	const std::string s;
};

//! CommandLineArguments class exceptions.
class CommandLineArgumentsException: public BaseException {
public:
	explicit CommandLineArgumentsException(const std::string &e) throw():
		BaseException(e) {}
};

//! Exceptions thrown if some network error occurs.
class NetworkException: public BaseException {
public:
	explicit NetworkException(const std::string &e) throw():
		BaseException(e) {}
};

//! Exceptions thrown if some runtime error occurs.
class RuntimeException: public BaseException {
public:
	explicit RuntimeException(const std::string &e) throw():
		BaseException(e) {}
};

//! Represents parsed command line arguments.
/*!
	This class is used to parse command line arguments.

	If an error occurs,
	CommandLineArgumentsException will be thrown. All other exceptions
	are propagated through the calls to the user of this class.

	It provides value semantics (objects copying allowed).
*/
class CommandLineArguments {
public:
					CommandLineArguments(int argc, const char **argv);
					~CommandLineArguments();

					// Copy ctor and assignment operator
					// generated by compiler are sufficient

	unsigned int	localPort() const { return _localPort; }
	std::string		hostAddress() const { return _hostAddress; }
	unsigned int	hostPort() const { return _hostPort; }

	static bool 	helpRequested(int argc, const char **argv);

private:
	void			parsePortNumber(const std::string &inputPortNumber,
									unsigned int &outputPortNumber);
	void			parseHostData(const std::string &rawHostData);
	static bool		isValidPortNumber(unsigned int portNumber);

private:
	unsigned int	_localPort;
	std::string		_hostAddress;
	unsigned int	_hostPort;

private:
	static const int			MandatoryArgumentsCount = 3;
	static const int			LocalPortArgumentIndex = 1;
	static const int			HostDataArgumentIndex = 2;
	static const unsigned int	DefaultHostPortNumber = 80;
	static const unsigned int	MinimalPortNumber = 1;
	static const unsigned int	MaximalPortNumber = 65536;
};

//! Base class for all HTTP messages.
/*!
	This class represents generic HTTP message and all specific HTTP message-like
	classes are derived from this class.
*/
class HttpMessage {
public:
	static const char	*HttpVersion10Marker;
	static const char	*HttpVersion11Marker;
	static const char	*HeaderValueDelimiter;
	static const char	*HeaderAndBodyDelimiter;
	static const char	*HeaderEndOfLineMarker;
	static const char	*ContentLengthHeaderName;

public:
	std::string			rawMessage() const;
	std::string			rawMessageBody() const;
	std::string			httpVersion() const;
	std::string			getHeaderValue(const std::string &headerName) const;
	bool				changeHeaderValue(const std::string &headerName,
										  const std::string &newHeaderValue);
	bool				connectionHasToBePersistent() const;
	virtual	bool		isComplete() const = 0;

protected:
	explicit			HttpMessage(std::string message = std::string(""));
	virtual				~HttpMessage();

						// Copy ctor and assignment operator
						// generated by compiler are sufficient

	virtual bool 		matchesHttpVersion(const std::string &message,
										   const std::string &httpVersionMarker) const = 0;
	bool				hasContentLengthHeader() const;
	bool				isCompleteByContentLength() const;

protected:
	std::string			_message;
};

//! This class represents HTTP requests.
/*!
	It provides value semantics (objects copying allowed).
*/
class HttpRequest: public HttpMessage {
public:
	static const char	*HttpRequestTypeGet;
	static const char	*DefaultObjectFileName;
	static const char	*HostHeaderName;

public:
	explicit			HttpRequest(std::string requestMessage = std::string(""));
						~HttpRequest();

						// Copy ctor and assignment operator
						// generated by compiler are sufficient

	std::string			httpRequestType() const;
	std::string			objectToGetfileName() const;
	virtual bool		isComplete() const;

	static HttpRequest	getSingleHttpRequest(std::string &rawData);

private:
	virtual bool 		matchesHttpVersion(const std::string &requestMessage,
										   const std::string &httpVersionMarker) const;
};

//! HTTP response.
/*!
	This class represents HTTP responses.

	It provides value semantics (objects copying allowed).
*/
class HttpResponse: public HttpMessage {
public:
	static const char			*TransferEncodingHeaderName;
	static const char			*TransferEncodingChunkedValue;
	static const char 			*TransferEncodingChunkedEndMarker;
	static const unsigned int	ResponseCodeStartIndex = 9;
	static const unsigned int	ResponseCodeLength = 3;

public:
	explicit			HttpResponse(std::string responseMessage = std::string(""));
						~HttpResponse();

						// Copy ctor and assignment operator
						// generated by compiler are sufficient

	virtual bool		isComplete() const;
	unsigned int		responseCode() const;

	static HttpResponse	createHttpNotFoundResponse(const std::string &httpVersion);

private:
	virtual bool 		matchesHttpVersion(const std::string &message,
										   const std::string &httpVersionMarker) const;
	bool				isCompleteByTransferEncodingChunked() const;
	bool				hasTransferEncodingChunkedHeader() const;
};

//! Base class for all connection classes.
/*!
	Every connection class is derived from this class.
*/
class Connection {
public:
						Connection();
	virtual				~Connection();

	virtual void		closeConnection() = 0;

private:
	// Disable copy ctor and assignment operator
	// since this is a reference object
	Connection(const Connection &);
	Connection & operator=(const Connection &);
};

//! Connection to a remote host.
/*!
	It provides reference semantics (objects copying is NOT allowed).
*/
class ConnectionToHost: public Connection {
public:
						ConnectionToHost(std::string hostAddress,
										 unsigned int hostPort);
	virtual				~ConnectionToHost();

	std::string			hostAddress() const;
	unsigned int		hostPort() const;
	virtual void		closeConnection();

	void				sendRequest(const HttpRequest &request);
	HttpResponse		receiveResponse();

private:
	// Disable copy ctor and assignment operator
	// since this is a reference object
	ConnectionToHost(const ConnectionToHost &);
	ConnectionToHost & operator=(const ConnectionToHost &);

	void				connectToHost();
	void				reconnectToHost();

private:
	std::string			_hostAddress;
	unsigned int		_hostPort;
	int					_hostSocket;
	sockaddr_in			_hostSocketInfo;

	static const int	BufferSize = 4096;
	char				buffer[BufferSize];
};

//! Connection from a client.
/*!
	It provides reference semantics (objects copying is NOT allowed).
*/
class ConnectionFromClient: public Connection {
public:
						ConnectionFromClient(int clientSocket,
											 sockaddr_in clientSocketInfo);
	virtual				~ConnectionFromClient();

	std::string			clientAddress() const;
	unsigned int		clientPort() const;
	virtual void		closeConnection();

	void				sendResponse(const HttpResponse &response);
	HttpRequest			receiveRequest();

private:
	// Disable copy ctor and assignment operator
	// since this is a reference object
	ConnectionFromClient(const ConnectionFromClient &);
	ConnectionFromClient & operator=(const ConnectionFromClient &);

private:
	int					_clientSocket;
	sockaddr_in 		_clientSocketInfo;

	static const int	BufferSize = 4096;
	char				buffer[BufferSize];
	std::string			rawData;
};

//! HTTP proxy server.
/*!
	It provides reference semantics (objects copying is NOT allowed).
*/
class Server {
public:
	explicit			Server(unsigned int localPort);
						~Server();

	std::auto_ptr<ConnectionToHost> connectToHost(std::string hostAddress,
												  unsigned int hostPort);
	std::auto_ptr<ConnectionFromClient> acceptConnection();
	void				processConnectionFromClient(
							std::auto_ptr<ConnectionFromClient> connectionFromClient,
							const std::string &hostAddress,
							unsigned int hostPort);
	void				loadForbiddenInfoList(const std::string &rawForbiddenInfo);
	bool				isForbidden(const std::string &clientAddress,
									const std::string &objectFileName);
	bool				processHttpRequest(ConnectionToHost *connectionToHost,
										   ConnectionFromClient *connectionFromClient,
										   HttpRequest &httpRequest);

	static std::string	getForbiddenInfoFileName() { return ForbiddenInfoFileName;}

private:
	// Disable copy ctor and assignment operator
	// since this is a reference object
	Server(const Server &);
	Server & operator=(const Server &);

	void 				setSignalHandling();
	void				restoreSignalHandling();
	static void			printRequestInfo(std::ostream &outputStream,
										 const std::string &clientAddress,
										 const std::string &objectFileName);

private:
	unsigned int		_localPort;
	int					_localSocket;
	sockaddr_in			_localSocketInfo;

	// Blocking masks
	sigset_t 			blockMaskCurrent;
	sigset_t 			blockMaskOld; // (for restoration)

	// Handlers
	struct sigaction	sigchldHandler;
	struct sigaction	sigchldHandlerOld; // (for restoration)

	typedef std::pair<std::string, std::string> ForbiddenInfo;
	typedef std::vector<ForbiddenInfo> ForbiddenInfoList;
	ForbiddenInfoList	_forbiddenInfoList;

private:
	static const int	MaxConnectionCount = 10;
	static const char	ForbiddenInfoRecordsDelimiter = '\n';
	static const char	ForbiddenInfoRecordPartsDelimiter = ':';
	static const char	*ForbiddenInfoFileName;
};

} // namespace ProxyHTTP

#endif // #ifndef PROXYHTTP_H

/* End of file proxyhttp.h */
